/**
 * Copyright (c) 1993 Richard V. Nash.
 * Copyright (c) 2000 Dan Papasian.  All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "includes.h"

static int check_kings(game_state_t *gs);
static int get_empty_slot(void);
static int WriteMoves(FILE *fp, move_t *m);
static int ReadMove(FILE *fp, move_t *m);
static void WriteGameState(FILE *fp, game_state_t *gs);
static int ReadGameState(FILE *fp, game_state_t *gs, int version);
static int ReadGameAttrs_common(FILE *fp, char *fname, int g, int version);
static long OldestHistGame(char *login);
static void RemoveHistGame(char *file, int maxlines);
static void write_g_out(int g, char *file, int maxlines, int isDraw, char *EndSymbol, char *name, time_t *now);
static int game_zero(int g);
 
const char *TypeStrings[NUM_GAMETYPES] = {"untimed", "blitz", "standard", 
					  "nonstandard", "wild", "lightning", 
					  "bughouse"};

const char *TypeChars[NUM_GAMETYPES] = {"U", "B", "S", "N", "W", "L", "B"};

static int get_empty_slot()
/* this method is awful! how about allocation as we need it and freeing
    afterwards! */
{
  int i;

  for (i = 0; i < game_globals.g_num; i++) {
    if (game_globals.garray[i].status == GAME_EMPTY)
      return i;
  }
  game_globals.g_num++;
  if (!game_globals.garray) {
    game_globals.garray = (game *) malloc(sizeof(game) * game_globals.g_num);
  } else {
    game_globals.garray = (game *) realloc(game_globals.garray, sizeof(game) * game_globals.g_num);
  } /* yeah great, bet this causes lag!  - DAV*/
    /* I have serious doubt of the truth to the above client- bugg */
  game_globals.garray[game_globals.g_num - 1].status = GAME_EMPTY;
  return game_globals.g_num - 1;
}

int game_new()
{
  int new = get_empty_slot();
  game_zero(new);
  return new;
}

static int game_zero(int g)
{
  game_globals.garray[g].white = -1;
  game_globals.garray[g].black = -1;

  game_globals.garray[g].status = GAME_NEW;
  game_globals.garray[g].link = -1;
  game_globals.garray[g].rated = 0;
  game_globals.garray[g].private = 0;
  game_globals.garray[g].result = END_NOTENDED;
  game_globals.garray[g].type = TYPE_UNTIMED;
  game_globals.garray[g].passes = 0;
  game_globals.garray[g].game_state.gameNum = g;
  game_globals.garray[g].numHalfMoves = 0;
  game_globals.garray[g].moveListSize = 0;
  game_globals.garray[g].moveList = NULL;
  game_globals.garray[g].examMoveListSize = 0;
  game_globals.garray[g].examMoveList = NULL;
  game_globals.garray[g].wInitTime = 300;	/* 5 minutes */
  game_globals.garray[g].wIncrement = 0;
  game_globals.garray[g].bInitTime = 300;	/* 5 minutes */
  game_globals.garray[g].bIncrement = 0;
  game_globals.garray[g].flag_pending = FLAG_NONE;
  game_globals.garray[g].flag_check_time = 0L;
  game_globals.garray[g].white_name[0] = '\0';
  game_globals.garray[g].black_name[0] = '\0';
  game_globals.garray[g].white_rating = 0;
  game_globals.garray[g].black_rating = 0;
  game_globals.garray[g].revertHalfMove = 0;
  strcpy (game_globals.garray[g].FENstartPos,"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w");
  return 0;
}

int game_free(int g)
{
  if (game_globals.garray[g].moveListSize)
    free(game_globals.garray[g].moveList);
  if (game_globals.garray[g].examMoveListSize)
    free(game_globals.garray[g].examMoveList);
  game_globals.garray[g].moveList = NULL;
  game_globals.garray[g].examMoveList = NULL;
  game_globals.garray[g].moveListSize = 0;
  game_globals.garray[g].examMoveListSize = 0;
  return 0;
}

int game_remove(int g)
{
  /* Should remove game from players observation list */
  game_free(g);
  game_zero(g);
  game_globals.garray[g].status = GAME_EMPTY;
  return 0;
}

/* old moves not stored now - uses smoves */
int game_finish(int g)
{
  player_game_ended(g);		/* Alert playerdb that game ended */
  game_remove(g);
  return 0;
}

void MakeFENpos (int g, char *FEN)
{
  strcpy(FEN, boardToFEN(g));
}

static char *game_time_str(int wt, int winc, int bt, int binc)
{
  static char tstr[50];

  if ((!wt) && (!winc)) {			/* Untimed */
    strcpy(tstr, "");
    return tstr;
  }
  if ((wt == bt) && (winc == binc)) {
    sprintf(tstr, " %d %d", wt, winc);
  } else {
    sprintf(tstr, " %d %d : %d %d", wt, winc, bt, binc);
  }
  return tstr;
}

const char *bstr[] = {"untimed", "blitz", "standard", "non-standard", "wild", "lightning", "Bughouse"};

const char *rstr[] = {"unrated", "rated"};

char *game_str(int rated, int wt, int winc, int bt, int binc,
		       char *cat, char *board)
{
  static char tstr[200];

  if (cat && cat[0] && board && board[0] &&
      (strcmp(cat, "standard") || strcmp(board, "standard"))) {
    sprintf(tstr, "%s %s%s Loaded from %s/%s",
	    rstr[rated],
	    bstr[game_isblitz(wt / 60, winc, bt / 60, binc, cat, board)],
	    game_time_str(wt / 60, winc, bt / 60, binc),
	    cat, board);
  } else {
    sprintf(tstr, "%s %s%s",
	    rstr[rated],
	    bstr[game_isblitz(wt / 60, winc, bt / 60, binc, cat, board)],
	    game_time_str(wt / 60, winc, bt / 60, binc));
  }
  return tstr;
}

int game_isblitz(int wt, int winc, int bt, int binc,
			 char *cat, char *board)
{
  int total;

  if(cat && cat[0]) {
    if (!strcmp(cat, "bughouse"))
      return TYPE_BUGHOUSE;
    if (board && board[0]) {
      if (!strcmp(cat, "wild"))
        return TYPE_WILD;
      if (strcmp(cat, "standard") || strcmp(board, "standard"))
        return TYPE_NONSTANDARD;
    }
  }

  if ((wt == 0) || (bt == 0))
			/* nonsense if one is timed and one is not */
    return TYPE_UNTIMED;

  if ((wt != bt) || (winc != binc))
    return TYPE_NONSTANDARD;
  total = wt * 60 + winc * 40;
  if (total < 180)		/* 3 minute */
    return TYPE_LIGHT;
  if (total >= 900)		/* 15 minutes */
    return TYPE_STAND;
  else
    return TYPE_BLITZ;
}

void send_board_to(int g, int p)
{
  char *b;
  int side;
  int relation;

/* since we know g and p, figure out our relationship to this game */

  side = WHITE;
  if ((game_globals.garray[g].status == GAME_EXAMINE) || (game_globals.garray[g].status == GAME_SETUP)) {
    if (player_globals.parray[p].game == g) {
      relation = 2;
    } else {
      relation = -2;
    }
  } else {
    if (player_globals.parray[p].game == g) {
      side = player_globals.parray[p].side;
      relation = ((side == game_globals.garray[g].game_state.onMove) ? 1 : -1);
    } else {
      relation = 0;
    }
  }

  if (CheckPFlag(p, PFLAG_FLIP)) {		/* flip board? */
    if (side == WHITE)
      side = BLACK;
    else
      side = WHITE;
  }
  game_update_time(g);
  b = board_to_string(game_globals.garray[g].white_name,
		      game_globals.garray[g].black_name,
		      game_globals.garray[g].wTime,
		      game_globals.garray[g].bTime,
		      &game_globals.garray[g].game_state,
		      (game_globals.garray[g].status == GAME_EXAMINE || game_globals.garray[g].status == GAME_SETUP) ?
		      game_globals.garray[g].examMoveList : game_globals.garray[g].moveList,
		      player_globals.parray[p].style,
		      side, relation, p);
  Bell(p);

  if (player_globals.parray[p].game == g && net_globals.con[player_globals.parray[p].socket]->timeseal) {
    pprintf_noformat(p, "\n%s\n[G]\n", b);
  } else {
    pprintf_noformat(p, "\n%s", b);
  }

  if (p != command_globals.commanding_player) {
	  send_prompt(p);
  }
}

void send_boards(int g)
{
  int p,p1;
  simul_info_t *simInfo = player_globals.parray[game_globals.garray[g].white].simul_info;
  int which_board = -1;

  /* Begin code added 3/28/96 - Figbert */

  if ((game_globals.garray[g].status != GAME_SETUP) && 
      (check_kings(&game_globals.garray[g].game_state))) {
    d_printf( "Game has invalid amount of kings.  Aborting...\n");
    game_finish(g);
    
    for (p = 0; p < player_globals.p_num; p++) {
      if (player_globals.parray[p].status == PLAYER_EMPTY)
        continue;
      if (player_globals.parray[p].game == g) {

        p1 = player_globals.parray[p].opponent;

        if (p1 >= 0 && player_globals.parray[p1].game == g) {
          pprintf (p1, "Disconnecting you from game number %d.\n", g+1);
          player_globals.parray[p1].game = -1;
        }

        pprintf (p, "Disconnecting you from game number %d.\n", g+1);
        player_globals.parray[p].game = -1;
      }
    }
    return;
  }

  /* End code added 3/28/96 - Figbert */
 
  if (simInfo != NULL)
    which_board = simInfo->boards[simInfo->onBoard];

  if ((simInfo == NULL) || (which_board == g)) {
    for (p = 0; p < player_globals.p_num; p++) {
      if (player_globals.parray[p].status == PLAYER_EMPTY)
        continue;
      if (player_is_observe(p, g) || (player_globals.parray[p].game == g))
        send_board_to(g, p);
    }
  }
}

void game_update_time(int g)
{
  unsigned now, timesince;

  if (game_globals.garray[g].clockStopped)
    return;
  if (game_globals.garray[g].type == TYPE_UNTIMED)
    return;
  now = tenth_secs();
  timesince = now - game_globals.garray[g].lastDecTime;
  if (game_globals.garray[g].game_state.onMove == WHITE) {
    game_globals.garray[g].wTime -= timesince;
  } else {
    game_globals.garray[g].bTime -= timesince;
  }
  game_globals.garray[g].lastDecTime = now;
}

static void game_update_times()
{
  int g;

  for (g = 0; g < game_globals.g_num; g++) {
    if (game_globals.garray[g].status != GAME_ACTIVE)
      continue;
    if (game_globals.garray[g].clockStopped)
      continue;
    game_update_time(g);
  }
}

static char *EndString(int g, int personal)
{
  static char endstr[200];
  char *blackguy, *whiteguy;
  static char blackstr[] = "Black";
  static char whitestr[] = "White";

  blackguy = (personal ? game_globals.garray[g].black_name : blackstr);
  whiteguy = (personal ? game_globals.garray[g].white_name : whitestr);

  switch (game_globals.garray[g].result) {
  case END_CHECKMATE:
    sprintf(endstr, "%s checkmated",
	    game_globals.garray[g].winner == WHITE ? blackguy : whiteguy);
    break;
  case END_RESIGN:
    sprintf(endstr, "%s resigned",
	    game_globals.garray[g].winner == WHITE ? blackguy : whiteguy);
    break;
  case END_FLAG:
    sprintf(endstr, "%s ran out of time",
	    game_globals.garray[g].winner == WHITE ? blackguy : whiteguy);
    break;
  case END_AGREEDDRAW:
    sprintf(endstr, "Game drawn by mutual agreement");
    break;
  case END_BOTHFLAG:
    sprintf(endstr, "Game drawn because both players ran out of time");
    break;
  case END_REPETITION:
    sprintf(endstr, "Game drawn by repetition");
    break;
  case END_50MOVERULE:
    sprintf(endstr, "Draw by the 50 move rule");
    break;
  case END_ADJOURN:
    sprintf(endstr, "Game adjourned by mutual agreement");
    break;
  case END_LOSTCONNECTION:
    sprintf(endstr, "%s lost connection, game adjourned",
	    game_globals.garray[g].winner == WHITE ? whiteguy : blackguy);
    break;
  case END_ABORT:
    sprintf(endstr, "Game aborted by mutual agreement");
    break;
  case END_STALEMATE:
    sprintf(endstr, "Stalemate.");
    break;
  case END_NOTENDED:
    sprintf(endstr, "Still in progress");
    break;
  case END_COURTESY:
    sprintf(endstr, "Game courtesyaborted by %s",
	    game_globals.garray[g].winner == WHITE ? whiteguy : blackguy);
    break;
  case END_COURTESYADJOURN:
    sprintf(endstr, "Game courtesyadjourned by %s",
	    game_globals.garray[g].winner == WHITE ? whiteguy : blackguy);
    break;
  case END_NOMATERIAL:
    sprintf(endstr, "Game drawn because neither player has mating material");
    break;
  case END_FLAGNOMATERIAL:
    sprintf(endstr, "%s ran out of time and %s has no material to mate",
	    game_globals.garray[g].winner == WHITE ? blackguy : whiteguy,
	    game_globals.garray[g].winner == WHITE ? whiteguy : blackguy);
    break;
  case END_ADJDRAW:
    sprintf(endstr, "Game drawn by adjudication");
    break;
  case END_ADJWIN:
    sprintf(endstr, "%s wins by adjudication",
	    game_globals.garray[g].winner == WHITE ? whiteguy : blackguy);
    break;
  case END_ADJABORT:
    sprintf(endstr, "Game aborted by adjudication");
    break;
  default:
    sprintf(endstr, "???????");
    break;
  }

  return (endstr);
}

char *EndSym(int g)
{
  static char *symbols[] = {"1-0", "0-1", "1/2-1/2", "*"};

  switch (game_globals.garray[g].result) {
  case END_CHECKMATE:
  case END_RESIGN:
  case END_FLAG:
  case END_ADJWIN:
    return ((game_globals.garray[g].winner == WHITE) ? symbols[0] : symbols[1]);
    break;
  case END_AGREEDDRAW:
  case END_BOTHFLAG:
  case END_REPETITION:
  case END_50MOVERULE:
  case END_STALEMATE:
  case END_NOMATERIAL:
  case END_FLAGNOMATERIAL:
  case END_ADJDRAW:
    return (symbols[2]);
    break;
  }

  return (symbols[3]);
}

/* This should be enough to hold any game up to at least 8000 moves
 * If we overwrite this, the server will crash :-).
 */
/* 8000? who you trying to kid? this is awful - enough for 600 halfs :) -DAV*/
#define GAME_STRING_LEN 19000
static char gameString[GAME_STRING_LEN];
char *movesToString(int g, int pgn)
{
  char tmp[160];
  int wr, br;
  int i, col;
  time_t curTime;
  move_t *moves;

  if (game_globals.garray[g].status == GAME_EXAMINE || game_globals.garray[g].status == GAME_SETUP)
    moves = game_globals.garray[g].examMoveList;
  else moves = game_globals.garray[g].moveList;

  wr = game_globals.garray[g].white_rating;
  br = game_globals.garray[g].black_rating;
  

  curTime = untenths(game_globals.garray[g].timeOfStart);

  if (pgn) {
    sprintf(gameString,
	    "\n[Event \"%s %s %s game\"]\n"
	    "[Site \"%s, %s\"]\n",
	    config_get_tmp("SERVER_NAME"),
	    rstr[game_globals.garray[g].rated], bstr[game_globals.garray[g].type],
	    config_get_tmp("SERVER_NAME"),
	    config_get_tmp("SERVER_LOCATION"));
    strftime(tmp, sizeof(tmp),
	     "[Date \"%Y.%m.%d\"]\n"
	     "[Time \"%H:%M:%S\"]\n",
	     localtime((time_t *) &curTime));
    strcat(gameString, tmp);
    sprintf(tmp,
	    "[Round \"-\"]\n"
	    "[White \"%s\"]\n"
	    "[Black \"%s\"]\n"
	    "[WhiteElo \"%d\"]\n"
	    "[BlackElo \"%d\"]\n",
	    game_globals.garray[g].white_name, game_globals.garray[g].black_name, wr, br);
    strcat(gameString, tmp);
    sprintf(tmp,
	    "[TimeControl \"%d+%d\"]\n"
	    "[Mode \"ICS\"]\n"
	    "[Result \"%s\"]\n\n",
	    game_globals.garray[g].wInitTime / 10, game_globals.garray[g].wIncrement / 10, EndSym(g));
    strcat(gameString, tmp);

    col = 0;
    for (i = 0; i < game_globals.garray[g].numHalfMoves; i++) {
      if (moves[i].color == WHITE) {
	if ((col += sprintf(tmp, "%d. ", (i+1) / 2 + 1)) > 70) {
	  strcat(gameString, "\n");
	  col = 0;
	}
	strcat(gameString, tmp);
      } else if (i==0) {
        strcat (tmp, "1. ... ");
        col += 7;
      }
      if ((col += sprintf(tmp, "%s ", moves[i].algString)) > 70) {
	strcat(gameString, "\n");
	col = 0;
      }
      strcat(gameString, tmp);
    }
    strcat(gameString, "\n");

  } else {

    sprintf(gameString, "\n%s ", game_globals.garray[g].white_name);
    if (wr > 0) {
      sprintf(tmp, "(%d) ", wr);
    } else {
      sprintf(tmp, "(UNR) ");
    }
    strcat(gameString, tmp);
    sprintf(tmp, "vs. %s ", game_globals.garray[g].black_name);
    strcat(gameString, tmp);
    if (br > 0) {
      sprintf(tmp, "(%d) ", br);
    } else {
      sprintf(tmp, "(UNR) ");
    }
    strcat(gameString, tmp);
    strcat(gameString, "--- ");
    strcat(gameString, strltime(&curTime));
    if (game_globals.garray[g].rated) {
      strcat(gameString, "\nRated ");
    } else {
      strcat(gameString, "\nUnrated ");
    }
    strcat (gameString, TypeStrings[game_globals.garray[g].type]);
    strcat(gameString, " match, initial time: ");
    if ((game_globals.garray[g].bInitTime != game_globals.garray[g].wInitTime) || (game_globals.garray[g].wIncrement != game_globals.garray[g].bIncrement)) { /* different starting times */ 
      sprintf(tmp, "%d minutes, increment: %d seconds AND %d minutes, increment: %d seconds.\n\n", game_globals.garray[g].wInitTime / 600, game_globals.garray[g].wIncrement / 10, game_globals.garray[g].bInitTime / 600, game_globals.garray[g].bIncrement / 10);
    } else {
      sprintf(tmp, "%d minutes, increment: %d seconds.\n\n", game_globals.garray[g].wInitTime / 600, game_globals.garray[g].wIncrement / 10);
    }
    strcat(gameString, tmp);
    sprintf(tmp, "Move  %-19s%-19s\n", game_globals.garray[g].white_name, game_globals.garray[g].black_name);
    strcat(gameString, tmp);
    strcat(gameString, "----  ----------------   ----------------\n");

    for (i = 0; i < game_globals.garray[g].numHalfMoves; i += 2) {
      if (i==0 && (moves[i].color == BLACK)) {
	sprintf(tmp, "%3d.  %-16s   %-16s\n", (i+1)/2 + 1, "...",
                     move_and_time(&moves[i]));
        i--;
      } else if (i + 1 < game_globals.garray[g].numHalfMoves) {
	sprintf(tmp, "%3d.  %-16s   ", (i+1)/2 + 1, move_and_time(&moves[i]));
	strcat(gameString, tmp);
	sprintf(tmp, "%-16s\n", move_and_time(&moves[i+1]));
      } else {
	sprintf(tmp, "%3d.  %-16s\n", (i+1)/2 + 1, move_and_time(&moves[i]));
      }
      strcat(gameString, tmp);
      if (strlen(gameString) > GAME_STRING_LEN - 100) {	/* Bug out if getting
							   close to filling this
							   string */
	return gameString;
      }
    }

    strcat(gameString, "      ");
  }

  sprintf(tmp, "{%s} %s\n", EndString(g, 0), EndSym(g));
  strcat(gameString, tmp);

  return gameString;
}

void game_disconnect(int g, int p)
{
  game_ended(g, (game_globals.garray[g].white == p) ? WHITE : BLACK, END_LOSTCONNECTION);
}

int CharToPiece(char c)
{
  switch (c) {
  case 'P':
    return W_PAWN;
  case 'p':
    return B_PAWN;
  case 'N':
    return W_KNIGHT;
  case 'n':
    return B_KNIGHT;
  case 'B':
    return W_BISHOP;
  case 'b':
    return B_BISHOP;
  case 'R':
    return W_ROOK;
  case 'r':
    return B_ROOK;
  case 'Q':
    return W_QUEEN;
  case 'q':
    return B_QUEEN;
  case 'K':
    return W_KING;
  case 'k':
    return B_KING;
  default:
    return NOPIECE;
  }
}

char PieceToChar(int piece)
{
  switch (piece) {
    case W_PAWN:return 'P';
  case B_PAWN:
    return 'p';
  case W_KNIGHT:
    return 'N';
  case B_KNIGHT:
    return 'n';
  case W_BISHOP:
    return 'B';
  case B_BISHOP:
    return 'b';
  case W_ROOK:
    return 'R';
  case B_ROOK:
    return 'r';
  case W_QUEEN:
    return 'Q';
  case B_QUEEN:
    return 'q';
  case W_KING:
    return 'K';
  case B_KING:
    return 'k';
  default:
    return ' ';
  }
}

/* One line has everything on it */
static int WriteMoves(FILE * fp, move_t *m)
{
  unsigned long MoveInfo = (m->color == BLACK);
  int piece, castle;
  int useFile = 0, useRank = 0, check = 0;
  int i;

  castle = (m->moveString[0] == 'o');
  if (castle)
    piece = KING;
  else
    piece = piecetype(CharToPiece(m->moveString[0]));

  MoveInfo = (MoveInfo <<= 3) | piece;
  MoveInfo = (MoveInfo <<= 3) | m->fromFile;
  MoveInfo = (MoveInfo <<= 3) | m->fromRank;
  MoveInfo = (MoveInfo <<= 3) | m->toFile;
  MoveInfo = (MoveInfo <<= 3) | m->toRank;
  MoveInfo = (MoveInfo <<= 3) | (m->pieceCaptured & 7);
  MoveInfo = (MoveInfo <<= 3) | (m->piecePromotionTo & 7);
  MoveInfo = (MoveInfo <<= 1) | (m->enPassant != 0);

  /* Are we using from-file or from-rank in algString? */
  i = strlen(m->algString) - 1;
  if (m->algString[i] == '+') {
    check = 1;
    i--;
  }
  if (piece != PAWN && !castle) {
    i -= 2;
    if (i < 0)
      return -1;
    if (m->algString[i] == 'x')
      i--;
    if (i < 0)
      return -1;
    if (isdigit(m->algString[i])) {
      useRank = 2;
      i--;
    }
    if (i < 0)
      return -1;
    useFile = (islower(m->algString[i]) ? 4 : 0);
  }
  MoveInfo = (MoveInfo << 3) | useFile | useRank | check;
  fprintf(fp, "%lx %x %x\n", MoveInfo, m->tookTime, m->atTime);

  return 0;
}

static int ReadMove(FILE * fp, move_t *m)
{
  char line[MAX_GLINE_SIZE];
  fgets(line, MAX_GLINE_SIZE - 1, fp);
  if (sscanf(line, "%d %d %d %d %d %d %d %d %d \"%[^\"]\" \"%[^\"]\" %u %u\n",
	     &m->color, &m->fromFile, &m->fromRank, &m->toFile, &m->toRank,
     &m->pieceCaptured, &m->piecePromotionTo, &m->enPassant, &m->doublePawn,
	     m->moveString, m->algString, &m->atTime, &m->tookTime) != 13)
    return -1;
  return 0;
}

static void WriteGameState(FILE * fp, game_state_t *gs)
{
  int i, j;

  for (i = 0; i < 8; i++)
    for (j = 0; j < 8; j++) {
      fprintf(fp, "%c", PieceToChar(gs->board[i][j]));
    }
  fprintf(fp, "%d %d %d %d %d %d",
	  gs->wkmoved, gs->wqrmoved, gs->wkrmoved,
	  gs->bkmoved, gs->bqrmoved, gs->bkrmoved);
  for (i = 0; i < 8; i++)
    fprintf(fp, " %d %d", gs->ep_possible[0][i], gs->ep_possible[1][i]);
  fprintf(fp, " %d %d %d\n", gs->lastIrreversable, gs->onMove, gs->moveNum);
}

static int ReadGameState(FILE * fp, game_state_t *gs, int version)
{
  int i, j;
  char pieceChar;
  int wkmoved, wqrmoved, wkrmoved, bkmoved, bqrmoved, bkrmoved;

  if (version == 0) {
    for (i = 0; i < 8; i++)
      for (j = 0; j < 8; j++)
	if (fscanf(fp, "%d ", &gs->board[i][j]) != 1)
	  return -1;
  } else {
    for (i = 0; i < 8; i++)
      for (j = 0; j < 8; j++) {
	pieceChar = getc(fp);
	gs->board[i][j] = CharToPiece(pieceChar);
      }
  }
  if (fscanf(fp, "%d %d %d %d %d %d",
	     &wkmoved, &wqrmoved, &wkrmoved,
	     &bkmoved, &bqrmoved, &bkrmoved) != 6)
    return -1;
  gs->wkmoved = wkmoved;
  gs->wqrmoved = wqrmoved;
  gs->wkrmoved = wkrmoved;
  gs->bkmoved = bkmoved;
  gs->bqrmoved = bqrmoved;
  gs->bkrmoved = bkrmoved;
  for (i = 0; i < 8; i++)
    if (fscanf(fp, " %d %d", &gs->ep_possible[0][i], &gs->ep_possible[1][i]) != 2)
      return -1;
  if (fscanf(fp, " %d %d %d\n", &gs->lastIrreversable, &gs->onMove, &gs->moveNum) != 3)
    return -1;
  return 0;
}

static int got_attr_value(int g, char *attr, char *value, FILE * fp, char *file)
{
  int i;

  if (!strcmp(attr, "w_init:")) {
    game_globals.garray[g].wInitTime = atoi(value);
  } else if (!strcmp(attr, "w_inc:")) {
    game_globals.garray[g].wIncrement = atoi(value);
  } else if (!strcmp(attr, "b_init:")) {
    game_globals.garray[g].bInitTime = atoi(value);
  } else if (!strcmp(attr, "b_inc:")) {
    game_globals.garray[g].bIncrement = atoi(value);
  } else if (!strcmp(attr, "white_name:")) {
    strcpy(game_globals.garray[g].white_name, value);
  } else if (!strcmp(attr, "black_name:")) {
    strcpy(game_globals.garray[g].black_name, value);
  } else if (!strcmp(attr, "white_rating:")) {
    game_globals.garray[g].white_rating = atoi(value);
  } else if (!strcmp(attr, "black_rating:")) {
    game_globals.garray[g].black_rating = atoi(value);
  } else if (!strcmp(attr, "result:")) {
    game_globals.garray[g].result = atoi(value);
  } else if (!strcmp(attr, "timestart:")) {
    game_globals.garray[g].timeOfStart = atoi(value);
  } else if (!strcmp(attr, "w_time:")) {
    game_globals.garray[g].wTime = atoi(value);
  } else if (!strcmp(attr, "b_time:")) {
    game_globals.garray[g].bTime = atoi(value);
  } else if (!strcmp(attr, "clockstopped:")) {
    game_globals.garray[g].clockStopped = atoi(value);
  } else if (!strcmp(attr, "rated:")) {
    game_globals.garray[g].rated = atoi(value);
  } else if (!strcmp(attr, "private:")) {
    game_globals.garray[g].private = atoi(value);
  } else if (!strcmp(attr, "type:")) {
    game_globals.garray[g].type = atoi(value);
  } else if (!strcmp(attr, "halfmoves:")) {
    game_globals.garray[g].numHalfMoves = atoi(value);
    if (game_globals.garray[g].numHalfMoves == 0)
      return 0;
    game_globals.garray[g].moveListSize = game_globals.garray[g].numHalfMoves;
    game_globals.garray[g].moveList = (move_t *) malloc(sizeof(move_t) * game_globals.garray[g].moveListSize);
    for (i = 0; i < game_globals.garray[g].numHalfMoves; i++) {
      if (ReadMove(fp, &game_globals.garray[g].moveList[i])) {
	d_printf( "CHESSD: Trouble reading moves from %s.\n", file);
	return -1;
      }
    }
  } else if (!strcmp(attr, "gamestate:")) {	/* Value meaningless */
    if (game_globals.garray[g].status != GAME_EXAMINE && game_globals.garray[g].status != GAME_SETUP &&
	ReadGameState(fp, &game_globals.garray[g].game_state, 0)) {
      d_printf( "CHESSD: Trouble reading game state from %s.\n", file);
      return -1;
    }
  } else {
    d_printf( "CHESSD: Error bad attribute >%s< from file %s\n", attr, file);
  }
  return 0;
}

static void ReadOneV1Move(FILE * fp, move_t *m)
{
  int i;
  char PieceChar;
  int useFile, useRank, check, piece;
  unsigned long MoveInfo;

  fscanf(fp, "%lx %x %x", &MoveInfo, &m->tookTime, &m->atTime);
  check = MoveInfo & 1;
  useRank = MoveInfo & 2;
  useFile = MoveInfo & 4;
  MoveInfo >>= 3;
  m->enPassant = MoveInfo & 1;	/* may have to negate later. */
  MoveInfo >>= 1;
  m->piecePromotionTo = MoveInfo & 7;	/* may have to change color. */
  MoveInfo >>= 3;
  m->pieceCaptured = MoveInfo & 7;	/* may have to change color. */
  MoveInfo >>= 3;
  m->toRank = MoveInfo & 7;
  MoveInfo >>= 3;
  m->toFile = MoveInfo & 7;
  MoveInfo >>= 3;
  m->fromRank = MoveInfo & 7;
  MoveInfo >>= 3;
  m->fromFile = MoveInfo & 7;
  MoveInfo >>= 3;
  piece = MoveInfo & 7;

  m->color = (MoveInfo & 8) ? BLACK : WHITE;
  if (m->pieceCaptured != NOPIECE) {
    if (m->color == BLACK)
      m->pieceCaptured |= WHITE;
    else
      m->pieceCaptured |= BLACK;
  }
  if (piece == PAWN) {
    PieceChar = 'P';
    if ((m->toRank == 3 && m->fromRank == 1)
	|| (m->toRank == 4 && m->fromRank == 6))
      m->doublePawn = m->toFile;
    else
      m->doublePawn = -1;
    if (m->pieceCaptured)
      sprintf(m->algString, "%cx%c%d", 'a' + m->fromFile,
	      'a' + m->toFile, m->toRank + 1);
    else
      sprintf(m->algString, "%c%d", 'a' + m->toFile, m->toRank + 1);
    if (m->piecePromotionTo != 0) {
      if (m->piecePromotionTo == KNIGHT)
	strcat(m->algString, "=N");
      else if (m->piecePromotionTo == BISHOP)
	strcat(m->algString, "=B");
      else if (m->piecePromotionTo == ROOK)
	strcat(m->algString, "=R");
      else if (m->piecePromotionTo == QUEEN)
	strcat(m->algString, "=Q");
      m->piecePromotionTo |= m->color;
    }
    if (m->enPassant)
      m->enPassant = m->toFile - m->fromFile;
  } else {
    m->doublePawn = -1;
    PieceChar = PieceToChar(piecetype(piece) | WHITE);
    if (PieceChar == 'K' && m->fromFile == 4 && m->toFile == 6) {
      strcpy(m->algString, "O-O");
      strcpy(m->moveString, "o-o");
    } else if (PieceChar == 'K' && m->fromFile == 4 && m->toFile == 2) {
      strcpy(m->algString, "O-O-O");
      strcpy(m->moveString, "o-o-o");
    } else {
      i = 0;
      m->algString[i++] = PieceChar;
      if (useFile)
	m->algString[i++] = 'a' + m->fromFile;
      if (useRank)
	m->algString[i++] = '1' + m->fromRank;
      if (m->pieceCaptured != 0)
	m->algString[i++] = 'x';
      m->algString[i++] = 'a' + m->toFile;
      m->algString[i++] = '1' + m->toRank;
      m->algString[i] = '\0';
    }
  }
  if (m->algString[0] != 'O')
    sprintf(m->moveString, "%c/%c%d-%c%d", PieceChar, 'a' + m->fromFile,
	    m->fromRank + 1, 'a' + m->toFile, m->toRank + 1);
  if (check)
    strcat(m->algString, "+");
}

static int ReadV1Moves(game *g, FILE * fp)
{
  int i;

  g->moveListSize = g->numHalfMoves;
  g->moveList = (move_t *) malloc(sizeof(move_t) * g->moveListSize);
  for (i = 0; i < g->numHalfMoves; i++) {
    ReadOneV1Move(fp, &g->moveList[i]);
  }
  return 0;
}

static int ReadV1GameFmt(struct game *g, FILE * fp, char *file, int version)
{
  char* FEN;
  char tmp[MAX_STRING_LENGTH];

  fscanf(fp, "%s %s", g->white_name, g->black_name);
  fscanf(fp, "%d %d", &g->white_rating, &g->black_rating);
  fscanf(fp, "%d %d %d %d", &g->wInitTime, &g->wIncrement,
	 &g->bInitTime, &g->bIncrement);
  if ((version < 3) && (!(g->bInitTime)))
    g->bInitTime = g->wInitTime;
                       /*PRE-V3 assumed bInitTime was 0 if balanced clocks*/
  fscanf(fp, "%lx", &g->timeOfStart);
  fscanf(fp, "%d %d", &g->wTime, &g->bTime);

/* fixing an (apparently) old bug: winner not saved */
  if (version > 1)
    fscanf(fp, "%d %d", &g->result, &g->winner);
  else
    fscanf(fp, "%d", &g->result);

  fscanf(fp, "%d %d %d %d", &g->private, (int *) &g->type,
	 &g->rated, &g->clockStopped);
  fscanf(fp, "%d", &g->numHalfMoves);
  ReadV1Moves(g, fp);

  if (version >= 4) {
    getc(fp);                   /* Skip past a newline. */ 

    fgets(tmp, MAX_LINE_SIZE, fp);
    tmp [strlen(tmp)-1] = '\0'; /* kill the newline char */

    if ((tmp[0] == '\0') || (!strcmp(tmp,"d w"))) /* default position */
      strcpy (g->FENstartPos,"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w");
    else {
      if (!strcmp(tmp,"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR"))
                 /* missing colour for default pos 1.7.1 bug */
        strcpy (g->FENstartPos,"rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w");
      else {
        strcpy (g->FENstartPos,tmp);
        FEN = g->FENstartPos;
        while (*(FEN) != '/') { /* check for missing fen pos (1.7.1 bug) */
          if (*(FEN++) == '\0') {
            fprintf (stderr,"Corrupt game %s vs %s!\n",g->white_name,g->black_name);
            return -1;
          }
        }
      }
    }
  } else
    getc(fp);                   /* Skip past a newline. */

  if (g->status != GAME_EXAMINE && g->status != GAME_SETUP) { 
    if (ReadGameState(fp, &g->game_state, version)) {
      d_printf( "CHESSD: Trouble reading game state from %s.\n", file);
      return -1;
    }
  } else if (g->status == GAME_EXAMINE)
    FEN_to_board(g->FENstartPos, &g->game_state);
  return 0;
}

static int ReadGameAttrs_common(FILE * fp, char *fname, int g,int version)
{
  int len;
  char *attr, *value;
  char line[MAX_GLINE_SIZE];

  if (version > 0) {
    if ((ReadV1GameFmt(&game_globals.garray[g], fp, fname, version)) < 0)
      return -1;
  }
  /* Read the game file here */
  else
    do {
      if ((len = strlen(line)) <= 1) {
	fgets(line, MAX_GLINE_SIZE - 1, fp);
	continue;
      }
      line[len - 1] = '\0';
      attr = eatwhite(line);
      if (attr[0] == '#')
	continue;		/* Comment */
      value = eatword(attr);
      if (!*value) {
	d_printf( "CHESSD: Error reading file %s\n", fname);
	fgets(line, MAX_GLINE_SIZE - 1, fp);
	continue;
      }
      *value = '\0';
      value++;
      value = eatwhite(value);
      if (!*value) {
	d_printf( "CHESSD: Error reading file %s\n", fname);
	fgets(line, MAX_GLINE_SIZE - 1, fp);
	continue;
      }
      stolower(attr);
      if (got_attr_value(g, attr, value, fp, fname)) {
	return -1;
      }
      fgets(line, MAX_GLINE_SIZE - 1, fp);
    } while (!feof(fp)); 
  if (!(game_globals.garray[g].bInitTime))
     game_globals.garray[g].bInitTime = game_globals.garray[g].wInitTime;
  return 0;
}

int ReadGameAttrs_exam(FILE * fp, char *fname, int g)
{
  int version = 0;
  char line[MAX_GLINE_SIZE];

  fgets(line, MAX_GLINE_SIZE - 1, fp);

   if (line[0] == 'v') {
    sscanf(line, "%*c %d", &version);

    if (version < 5)
      return -1;
    else
      return ReadGameAttrs_common(fp,fname,g,version);
  }

   else return -1;
}

int ReadGameAttrs(FILE * fp, char *fname, int g)
{
	int version = 0;
	char line[MAX_GLINE_SIZE];

	fgets(line, MAX_GLINE_SIZE - 1, fp);

	if (line[0] == 'v') {
		sscanf(line, "%*c %d", &version);
	}
	return ReadGameAttrs_common(fp,fname,g,version);
}

int game_read(int g, int wp, int bp)
{
  FILE *fp;
  char fname[MAX_FILENAME_SIZE];
  int piece;

  game_globals.garray[g].white = wp;
  game_globals.garray[g].black = bp;
/*  game_globals.garray[g].old_white = -1;
    game_globals.garray[g].old_black = -1;
*/
  game_globals.garray[g].moveListSize = 0;
  game_globals.garray[g].game_state.gameNum = g;
  strcpy(game_globals.garray[g].white_name, player_globals.parray[wp].name);
  strcpy(game_globals.garray[g].black_name, player_globals.parray[bp].name);
  if (game_globals.garray[g].type == TYPE_BLITZ) {
    game_globals.garray[g].white_rating = player_globals.parray[wp].b_stats.rating;
    game_globals.garray[g].black_rating = player_globals.parray[bp].b_stats.rating;
  } else if (game_globals.garray[g].type == TYPE_WILD) {
    game_globals.garray[g].white_rating = player_globals.parray[wp].w_stats.rating;
    game_globals.garray[g].black_rating = player_globals.parray[bp].w_stats.rating;
  } else if (game_globals.garray[g].type == TYPE_LIGHT) {
    game_globals.garray[g].white_rating = player_globals.parray[wp].l_stats.rating;
    game_globals.garray[g].black_rating = player_globals.parray[bp].l_stats.rating;
  } else if (game_globals.garray[g].type == TYPE_BUGHOUSE) {
    game_globals.garray[g].white_rating = player_globals.parray[wp].bug_stats.rating;
    game_globals.garray[g].black_rating = player_globals.parray[bp].bug_stats.rating;
  } else {
    game_globals.garray[g].white_rating = player_globals.parray[wp].s_stats.rating;
    game_globals.garray[g].black_rating = player_globals.parray[bp].s_stats.rating;
  }
  sprintf(fname, "%s/%c/%s-%s", ADJOURNED_DIR, player_globals.parray[wp].login[0],
	  player_globals.parray[wp].login, player_globals.parray[bp].login);
  fp = fopen(fname, "r");
  if (!fp) {
    return -1;
  }
  for (piece=PAWN; piece <= QUEEN; piece++) {
    game_globals.garray[g].game_state.holding[0][piece-PAWN]
      = game_globals.garray[g].game_state.holding[1][piece-PAWN] = 0;
  }
  if (ReadGameAttrs(fp, fname, g) < 0) {
    fclose(fp);
    return -1;
  }
  fclose(fp);

  if (game_globals.garray[g].result == END_ADJOURN
      || game_globals.garray[g].result == END_COURTESYADJOURN)
    game_globals.garray[g].result = END_NOTENDED;
  game_globals.garray[g].status = GAME_ACTIVE;
  game_globals.garray[g].startTime = tenth_secs();
  game_globals.garray[g].lastMoveTime = game_globals.garray[g].startTime;
  game_globals.garray[g].lastDecTime = game_globals.garray[g].startTime;
  /* Need to do notification and pending cleanup */
  return 0;
}

int game_delete(int wp, int bp)
{
  char fname[MAX_FILENAME_SIZE];
  char lname[MAX_FILENAME_SIZE];

  sprintf(fname, "%s/%c/%s-%s", ADJOURNED_DIR, player_globals.parray[wp].login[0],
	  player_globals.parray[wp].login, player_globals.parray[bp].login);
  sprintf(lname, "%s/%c/%s-%s", ADJOURNED_DIR, player_globals.parray[bp].login[0],
	  player_globals.parray[wp].login, player_globals.parray[bp].login);
  unlink(fname);
  unlink(lname);
  return 0;
}

static void WriteGameFile(FILE * fp, int g)
{
  int i;
  game *gg = &game_globals.garray[g];
  player *wp = &player_globals.parray[gg->white], *bp = &player_globals.parray[gg->black];

  fprintf(fp, "v %d\n", GAMEFILE_VERSION);
  fprintf(fp, "%s %s\n", wp->name, bp->name);
  fprintf(fp, "%d %d\n", gg->white_rating, gg->black_rating);
  fprintf(fp, "%d %d %d %d\n", gg->wInitTime, gg->wIncrement,
	  gg->bInitTime, gg->bIncrement);
  fprintf(fp, "%lx\n", gg->timeOfStart);
  fprintf(fp, "%d %d\n",
    (net_globals.con[wp->socket]->timeseal ? gg->wRealTime/100 : gg->wTime),
    (net_globals.con[bp->socket]->timeseal ? gg->bRealTime/100 : gg->bTime));
  fprintf(fp, "%d %d\n", gg->result, gg->winner);
  fprintf(fp, "%d %d %d %d\n", gg->private, gg->type,
	  gg->rated, gg->clockStopped);
  fprintf(fp, "%d\n", gg->numHalfMoves);
  for (i = 0; i < game_globals.garray[g].numHalfMoves; i++) {
    WriteMoves(fp, &game_globals.garray[g].moveList[i]);
  }
/* took out the next 3 lines to see if it helps with the crash bug we are
   having on examine...  fb 2.25.96 */
/* The next three lines stop wild games crashing the system - they are vital.
   I fixed the problem with examine - don't remove these again - DAV */
  if (strcmp (gg->FENstartPos, "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w"))
    fprintf (fp, "%s\n",gg->FENstartPos);
  else
    fprintf (fp, "d w\n");
  WriteGameState(fp, &game_globals.garray[g].game_state);
}

int game_save(int g)
{
  FILE *fp;
  player *wp, *bp;
  game *gg = &game_globals.garray[g];
  char fname[MAX_FILENAME_SIZE];
  char lname[MAX_FILENAME_SIZE];

  wp = &player_globals.parray[gg->white];
  bp = &player_globals.parray[gg->black];
  sprintf(fname, "%s/%c/%s-%s", ADJOURNED_DIR, wp->login[0],
	  wp->login, bp->login);
  sprintf(lname, "%s/%c/%s-%s", ADJOURNED_DIR, bp->login[0],
	  wp->login, bp->login);
  fp = fopen(fname, "w");
  if (!fp) {
    d_printf( "CHESSD: Problem opening file %s for write\n", fname);
    return -1;
  }
  WriteGameFile(fp, g);
  fclose(fp);
  /* Create link for easier stored game finding */
  if (bp->login[0] != wp->login[0])
    link(fname, lname);
  return 0;
}

static long OldestHistGame(char *login)
{
  FILE *fp;
  char pFile[MAX_FILENAME_SIZE];
  long when;

  sprintf(pFile, "%s/player_data/%c/%s.%s", STATS_DIR,
	  login[0], login, STATS_GAMES);
  fp = fopen(pFile, "r");

  if (fp == NULL) {
    sprintf(pFile, "%s/player_data/%c/.rem.%s.%s", STATS_DIR,
	    login[0], login, STATS_GAMES);
    fp = fopen(pFile, "r");
  }
  if (fp != NULL) {
    fscanf(fp, "%*d %*c %*d %*c %*d %*s %*s %*d %*d %*d %*d %*s %*s %ld",
	   &when);
    fclose(fp);
    return when;
  } else
    return 0L;
}

static void RemoveHistGame(char *file, int maxlines)
{
  FILE *fp;
  char GameFile[MAX_FILENAME_SIZE];
  char Opponent[MAX_LOGIN_NAME];
  char line[MAX_LINE_SIZE];
  long When, oppWhen;
  int count = 0;

  fp = fopen(file, "r");
  if (fp == NULL)
    return;

  fgets(line, MAX_LINE_SIZE - 1, fp);
  sscanf(line, "%*d %*c %*d %*c %*d %s %*s %*d %*d %*d %*d %*s %*s %ld",
	 Opponent, &When);
  count++;

  while (!feof(fp)) {
    fgets(line, MAX_LINE_SIZE - 1, fp);
    if (!feof(fp))
      count++;
  }
  fclose(fp);

  stolower(Opponent);
  if (count > maxlines) {
    truncate_file(file, maxlines);

    oppWhen = OldestHistGame(Opponent);
    if (oppWhen > When || oppWhen <= 0L) {
      sprintf(GameFile, "%s/%ld/%ld", HISTORY_DIR, When % 100, When);
      unlink(GameFile);
    }
  }
}

void RemHist(char *who)
{
  FILE *fp;
  char fName[MAX_FILENAME_SIZE];
  char Opp[MAX_LOGIN_NAME];
  long When, oppWhen;

  sprintf(fName, "%s/player_data/%c/%s.%s", STATS_DIR,
	  who[0], who, STATS_GAMES);
  fp = fopen(fName, "r");
  if (fp != NULL) {
    while (!feof(fp)) {
      fscanf(fp, "%*d %*c %*d %*c %*d %s %*s %*d %*d %*d %*d %*s %*s %ld",
	     Opp, &When);
      stolower(Opp);
      oppWhen = OldestHistGame(Opp);
      if (oppWhen > When || oppWhen <= 0L) {
	sprintf(fName, "%s/%ld/%ld", HISTORY_DIR, When % 100, When);
	unlink(fName);
      }
    }
    fclose (fp);
  }
}

static void write_g_out(int g, char *file, int maxlines, int isDraw,
			  char *EndSymbol, char *name, time_t *now)
{
  FILE *fp;
  int wp, bp;
  int wr, br;
  char type[4];
  char tmp[2048];
  char *ptmp = tmp;
  char cResult;
  int count = -1;
  char *goteco;

  wp = game_globals.garray[g].white;
  bp = game_globals.garray[g].black;

  if (game_globals.garray[g].private) {
    type[0] = 'p';
  } else {
    type[0] = ' ';
  }
  if (game_globals.garray[g].type == TYPE_BLITZ) {
    wr = player_globals.parray[wp].b_stats.rating;
    br = player_globals.parray[bp].b_stats.rating;
    type[1] = 'b';
  } else if (game_globals.garray[g].type == TYPE_WILD) {
    wr = player_globals.parray[wp].w_stats.rating;
    br = player_globals.parray[bp].w_stats.rating;
    type[1] = 'w';
  } else if (game_globals.garray[g].type == TYPE_STAND) {
    wr = player_globals.parray[wp].s_stats.rating;
    br = player_globals.parray[bp].s_stats.rating;
    type[1] = 's';
  } else if (game_globals.garray[g].type == TYPE_LIGHT) {
    wr = player_globals.parray[wp].l_stats.rating;
    br = player_globals.parray[bp].l_stats.rating;
    type[1] = 'l';
  } else if (game_globals.garray[g].type == TYPE_BUGHOUSE) {
    wr = player_globals.parray[wp].bug_stats.rating;
    br = player_globals.parray[bp].bug_stats.rating;
    type[1] = 'B';
  } else {
    wr = 0;
    br = 0;
    if (game_globals.garray[g].type == TYPE_NONSTANDARD)
      type[1] = 'n';
    else
      type[1] = 'u';
  }
  if (game_globals.garray[g].rated) {
    type[2] = 'r';
  } else {
    type[2] = 'u';
  }
  type[3] = '\0';

  fp = fopen(file, "r");
  if (fp) {
    while (!feof(fp))
      fgets(tmp, 1024, fp);
    sscanf(ptmp, "%d", &count);
    fclose(fp);
  }
  count = (count + 1) % 100;

  fp = fopen(file, "a");
  if (!fp)
    return;

  goteco = getECO(g);

/* Counter Result MyRating MyColor OppRating OppName [pbr 2 12 2 12] ECO End Date */
  if (name == player_globals.parray[wp].name) {
    if (isDraw)
      cResult = '=';
    else if (game_globals.garray[g].winner == WHITE)
      cResult = '+';
    else
      cResult = '-';

    fprintf(fp, "%d %c %d W %d %s %s %d %d %d %d %s %s %ld\n",
	    count, cResult, wr, br, player_globals.parray[bp].name, type,
	    game_globals.garray[g].wInitTime, game_globals.garray[g].wIncrement,
	    game_globals.garray[g].bInitTime, game_globals.garray[g].bIncrement,
	    goteco,
	    EndSymbol,
	    (long) *now);
  } else {
    if (isDraw)
      cResult = '=';
    else if (game_globals.garray[g].winner == BLACK)
      cResult = '+';
    else
      cResult = '-';

    fprintf(fp, "%d %c %d B %d %s %s %d %d %d %d %s %s %ld\n",
	    count, cResult, br, wr, player_globals.parray[wp].name, type,
	    game_globals.garray[g].wInitTime, game_globals.garray[g].wIncrement,
	    game_globals.garray[g].bInitTime, game_globals.garray[g].bIncrement,
	    goteco,
	    EndSymbol,
	    (long) *now);
  }
  fclose(fp);

  RemoveHistGame(file, maxlines);
}

/* Test if entry is present - 1 yes 0 no */
/* returns -1 if an error */

char get_journalgame_type(int p,char* fname,char slot)

{
  char cur_slot;
  char type[4];

  FILE* fp = fopen(fname,"r");
  if (!fp) {
    fprintf (stderr, "Corrupt journal file! %s\n",fname);
    pprintf (p, "The journal file is corrupt! See an admin.\n");
    return 0;
  }

  while (!feof(fp)) {
    if (fscanf(fp, "%c %*s %*d %*s %*d %s %*d %*d %*s %*s %*s\n",
       &cur_slot,type) != 2) {
      d_printf( "CHESSD: Error in journal info format for player %s.\n",
        player_globals.parray[p].name);
      pprintf(p, "The journal file is corrupt! Error in internal format.\n");
      return '\0';
    }
    if (cur_slot == slot) {
      fclose (fp);
      if (type[0] == 'p')
        return type[1];
      else
        return type[0];
    }
  }
  fclose (fp);
  return '0';
}


/* Returns 1 if successful */

int removejournalitem(int p, char slot,FILE* fp,char* fname, int* empty)

{

  FILE* fp_new;
  int found = 0;
  journal* j;

 *empty = 1;
 fp_new = fopen(fname, "w");
  if (!fp_new) {
    fprintf (stderr, "Can't write to journal %s!\n",fname);
    pprintf (p, "Was unable to write to the file! See an admin.\n");
    return 0;
  }
  j = (journal*) malloc(sizeof(journal));
  while (!feof(fp)) {
    if (fscanf(fp, "%c %s %d %s %d %s %d %d %s %s %s\n",
               &j->slot,
               j->WhiteName,
               &j->WhiteRating,
               j->BlackName,
               &j->BlackRating,
               j->type,
               &j->t, &j->i,
               j->eco,
               j->ending,
               j->result) != 11) {
      d_printf( "CHESSD: Error in journal info format. %s\n", fname);
      pprintf(p, "The journal file is corrupt! Error in internal format.\n");
      fclose(fp_new);
      free (j);
      return 0;
    }
    if (slot != j->slot) {
      *empty = 0;
      fprintf(fp_new, "%c %s %d %s %d %s %d %d %s %s %s\n",
               j->slot,
               j->WhiteName,
               j->WhiteRating,
               j->BlackName,
               j->BlackRating,
               j->type,
               j->t, j->i,
               j->eco,
               j->ending,
               j->result);
    } else
      found = 1;
  }
  fclose(fp_new);
  free (j);
  return found;
}

/* Find from_spot in journal list - return 0 if corrupted */

int journal_get_info(int p, char from_spot, journal* j, char *fname)
{
  FILE *fp;

  fp = fopen(fname, "r");
  if (!fp) {
    fprintf (stderr, "Corrupt journal file! %s\n",fname);
    pprintf (p, "The journal file is corrupt! See an admin.\n");
    return 0;
  }
  while (!feof(fp)) {
    if (fscanf(fp, "%c %s %d %s %d %s %d %d %s %s %s\n",
               &j->slot,
               j->WhiteName,
               &j->WhiteRating,
               j->BlackName,
               &j->BlackRating,
               j->type,
               &j->t, &j->i,
               j->eco,
               j->ending,
               j->result) != 11) {
      d_printf( "CHESSD: Error in journal info format. %s\n", fname);
      pprintf(p, "The journal file is corrupt! Error in internal format.\n");
      fclose(fp);
      return 0;
    }
    if (tolower(j->slot) == from_spot) {
       fclose(fp);
       return 1;
    }
  }
  fclose(fp);
  return 0;
}

void addjournalitem(int p, journal* j_add, char* fname)
{
  journal* j_cur;
  int have_output=0;
  char fname2[MAX_FILENAME_SIZE];

 FILE *fp;
 FILE *fp2;

  strcpy (fname2,fname);
  strcat (fname2,".w");
  fp2 = fopen(fname2, "w");
  if (!fp2) {
    d_printf( "CHESSD: Problem opening file %s for write\n", fname);
    pprintf (p, "Couldn't update journal! Report this to an admin.\n");
    return;
  } 
  fp = fopen(fname, "r");
  if (!fp) { /* Empty? */
    fprintf(fp2, "%c %s %d %s %d %s %d %d %s %s %s\n",
       j_add->slot, j_add->WhiteName, j_add->WhiteRating,
       j_add->BlackName, j_add->BlackRating,
       j_add->type, j_add->t, j_add->i, j_add->eco, j_add->ending,
       j_add->result);
    fclose (fp2);
    rename (fname2, fname);
    return;
  } else {
   j_cur = (journal*) malloc(sizeof(journal));
   while (!feof(fp)) {
    if (fscanf(fp, "%c %s %d %s %d %s %d %d %s %s %s\n",
               &j_cur->slot,
               j_cur->WhiteName,
               &j_cur->WhiteRating,
               j_cur->BlackName,
               &j_cur->BlackRating,
               j_cur->type,
               &j_cur->t, &j_cur->i,
               j_cur->eco,
               j_cur->ending,
               j_cur->result) != 11) {
               d_printf( "CHESSD: Error in journal info format - aborting. %s\n", fname);
               free (j_cur);
               fclose(fp);
               fclose(fp2);
               return;
               }
    if ((j_cur->slot >= j_add->slot) && (!have_output)) {

      fprintf(fp2, "%c %s %d %s %d %s %d %d %s %s %s\n",
             j_add->slot,
             j_add->WhiteName,
             j_add->WhiteRating,
             j_add->BlackName,
             j_add->BlackRating,
             j_add->type,
             j_add->t, j_add->i,
             j_add->eco,
             j_add->ending,
             j_add->result);
      have_output = 1;

    } 
    if (j_cur->slot != j_add->slot) {

      fprintf(fp2, "%c %s %d %s %d %s %d %d %s %s %s\n",
             j_cur->slot,
             j_cur->WhiteName,
             j_cur->WhiteRating,
             j_cur->BlackName,
             j_cur->BlackRating,
             j_cur->type,
             j_cur->t, j_cur->i,
             j_cur->eco,
             j_cur->ending,
             j_cur->result);
    }
  }

  if (!have_output) { /* Haven't written yet */
    fprintf(fp2, "%c %s %d %s %d %s %d %d %s %s %s\n",
                 j_add->slot,
                 j_add->WhiteName,
                 j_add->WhiteRating,
                 j_add->BlackName,
                 j_add->BlackRating,
                 j_add->type,
                 j_add->t, j_add->i,
                 j_add->eco,
                 j_add->ending,
                 j_add->result);
    }
  free (j_cur);
  fclose(fp);
  fclose(fp2);
  rename(fname2, fname);
  return;
  }
} 

int pjournal(int p, int p1, char *fname)
{
  FILE *fp;
  journal* j;

  fp = fopen(fname, "r");
  if (!fp) {
    pprintf(p, "Sorry, no journal information available.\n");
    return COM_OK;
  }

  j = (journal*) malloc(sizeof(journal));
  pprintf(p, "Journal for %s:\n", player_globals.parray[p1].name);
  pprintf(p, "   White         Rating  Black         Rating  Type         ECO End Result\n");
  while (!feof(fp)) {
    if (fscanf(fp, "%c %s %d %s %d %s %d %d %s %s %s\n",
               &j->slot,
               j->WhiteName,
               &j->WhiteRating,
               j->BlackName,
               &j->BlackRating,
               j->type,
               &j->t, &j->i,
               j->eco,
               j->ending,
               j->result) != 11) {
      d_printf( "CHESSD: Error in journal info format. %s\n", fname);
      fclose(fp);
      free(j);
      return COM_OK;
    }
    j->WhiteName[13] = '\0';         /* only first 13 chars in name */
    j->BlackName[13] = '\0';
    pprintf(p, "%c: %-13s %4d    %-13s %4d    [%3s%3d%4d] %s %3s %-7s\n",
            j->slot, j->WhiteName, j->WhiteRating,
            j->BlackName, j->BlackRating,
            j->type, j->t / 600, j->i / 10, j->eco, j->ending,
            j->result);
  }
  free (j);
  fclose(fp);
  return COM_OK;
}

int pgames(int p, int p1, char *fname)
{
  FILE *fp;
  time_t t;
  int MyRating, OppRating;
  int wt, wi, bt, bi;
  char OppName[MAX_LOGIN_NAME + 1];
  char type[100];
  char eco[100];
  char ending[100];
  char MyColor[2];
  int count;
  char result[2];

  fp = fopen(fname, "r");
  if (!fp) {
    pprintf(p, "Sorry, no game information available.\n");
    return COM_OK;
  }
  pprintf(p, "History for %s:\n", player_globals.parray[p1].name);
  pprintf(p, "                  Opponent      Type         ECO End Date\n");
  while (!feof(fp)) {
    if (fscanf(fp, "%d %s %d %s %d %s %s %d %d %d %d %s %s %ld\n",
	       &count,
	       result,
	       &MyRating,
	       MyColor,
	       &OppRating,
	       OppName,
	       type,
	       &wt, &wi,
	       &bt, &bi,
	       eco,
	       ending,
	       (long *) &t) != 14) {
      d_printf( "CHESSD: Error in games info format. %s\n", fname);
      fclose(fp);
      return COM_OK;
    }
    OppName[13] = '\0';		/* only first 13 chars in name */
    pprintf(p, "%2d: %s %4d %s %4d %-13s [%3s%3d%4d] %s %3s %s",
	    count, result, MyRating, MyColor,
	    OppRating, OppName,
	    type, wt / 600, wi / 10, eco, ending,
	    ctime(&t));
  }
  fclose(fp);
  return COM_OK;
}

void game_write_complete(int g, int isDraw, char *EndSymbol)
{
  char fname[MAX_FILENAME_SIZE];
  int wp = game_globals.garray[g].white, bp = game_globals.garray[g].black;
  time_t now = time(NULL);
  int Result;
  FILE *fp = NULL;

  do {
    if (fp) {
      fclose(fp);
      now++;
    }
    sprintf(fname, "%s/%ld/%ld", HISTORY_DIR, (long) now % 100, (long) now);
    fp = fopen(fname, "r");
  } while (fp);   /* terminates when the file doesn't exist */
 
  fp = fopen(fname, "w");

  if (fp) {
    WriteGameFile(fp, g);
    fclose(fp);
  } else {
    d_printf( "Trouble writing history file %s", fname);
  }

  sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR,
	  player_globals.parray[wp].login[0], player_globals.parray[wp].login, STATS_GAMES);
  write_g_out(g, fname, 10, isDraw, EndSymbol, player_globals.parray[wp].name, &now);
  sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR,
	  player_globals.parray[bp].login[0], player_globals.parray[bp].login, STATS_GAMES);
  write_g_out(g, fname, 10, isDraw, EndSymbol, player_globals.parray[bp].name, &now);

  if (isDraw)
    Result = -1;
  else if (game_globals.garray[g].winner == WHITE)
    Result = 1;
  else
    Result = 0;
  sprintf(fname,"%s/player_data/%c/%s.gstats", STATS_DIR, player_globals.parray[game_globals.garray[g].white].login[0], player_globals.parray[game_globals.garray[g].white].login);
  if ((CheckPFlag(bp, PFLAG_REG)) && (CheckPFlag(wp, PFLAG_REG)) && (game_globals.garray[g].type != TYPE_WILD))
  	game_save_playerratio(fname,player_globals.parray[game_globals.garray[g].black].name,Result,game_globals.garray[g].rated);
  if (isDraw)
    Result = -1;
  else if (game_globals.garray[g].winner == BLACK)
    Result = 1;
  else
    Result = 0;
  sprintf(fname,"%s/player_data/%c/%s.gstats", STATS_DIR, player_globals.parray[game_globals.garray[g].black].login[0], player_globals.parray[game_globals.garray[g].black].login);
  if ((CheckPFlag(bp, PFLAG_REG)) && (CheckPFlag(wp, PFLAG_REG)) && (game_globals.garray[g].type != TYPE_WILD))
  	game_save_playerratio(fname,player_globals.parray[game_globals.garray[g].white].name,Result,game_globals.garray[g].rated);

}

int game_count(void)
{
  int g, count = 0;

  for (g = 0; g < game_globals.g_num; g++) {
    if ((game_globals.garray[g].status == GAME_ACTIVE) || (game_globals.garray[g].status == GAME_EXAMINE)
          || (game_globals.garray[g].status == GAME_SETUP))
      count++;
  }
  if (count > command_globals.game_high)
    command_globals.game_high = count;
  return count;
}

static int check_kings(game_state_t *gs)
{
  /* Function written 3/28/96 by Figbert to check for 1 king each side only! */
  int blackking = 0, whiteking = 0;
  int f, r;


  for (f = 0; f < 8; f++)
    for (r = 0; r < 8; r++) {
      if (gs->board[f][r] == B_KING) blackking++;
      if (gs->board[f][r] == W_KING) whiteking++;
    }

  if (blackking == 1 && whiteking == 1) return 0; /* Perfect! */
  
  return -1;
}
