/*
   Copyright (C) Andrew Tridgell 2002
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  a malloc wrapper library for the dynamically loaded part of
  chessd. This is used both as a malloc checker library and to prevent
  some bad interactions between the dl library and malloc
*/


#include "includes.h"
#undef free
#undef strdup
#undef strndup

#define MAGIC1 0x28021967
#define MAGIC2 0x26011966

/*
  a magic-value protected malloc wrapper
*/
void *malloc(size_t size)
{
	void *(*fn)(size_t ) = dlsym((void *)-1, "malloc");
	void *ptr;

	ptr = fn(size+12);

#if MALLOC_DEBUG
	d_printf("malloc(%d) -> %p\n", size, ptr);
#endif

	if (!ptr) return NULL;

	memset(ptr, 0, size+12);

	((unsigned *)ptr)[0] = MAGIC1;
	((unsigned *)ptr)[1] = size;

	*(unsigned *)(ptr + size + 8) = MAGIC2;

	return ptr+8;
}

/*
  determine the size of a malloced region
*/
static size_t _malloc_size(void *ptr)
{
	if (!ptr) return 0;

	ptr -= 8;

	return ((unsigned *)ptr)[1];
}

/*
  a paranoid free() function that checks for simple corruption
*/
 void free(void *ptr)
{
	void (*fn)(void * ) = dlsym((void *)-1, "free");
	size_t size;
	unsigned v1, v2;

	if (!ptr) return;

	ptr -= 8;

	v1 = ((unsigned *)ptr)[0];
	size = ((unsigned *)ptr)[1];
	v2 = *(unsigned *)(ptr + size + 8);

	if (v1 != MAGIC1 || v2 != MAGIC2) {
		d_printf("malloc corruption ptr=%p v1=%x v2=%x size=%d\n", 
			ptr, v1, v2, size);
		return;
	}

#if MALLOC_DEBUG
	d_printf("free(%p) size=%d\n", ptr, size);
#endif

	fn(ptr);
}

/* a calloc wrapper */
void *calloc(size_t n, size_t size)
{
	void *ptr;
	ptr = malloc(n*size);
	if (ptr) memset(ptr, 0, n*size);
	return ptr;
}

/* strdup wrapper */
 char *strdup(const char *s)
{
	int len = strlen(s);
	char *ret;
	ret = malloc(len+1);
	memcpy(ret, s, len+1);
	return ret;
}

/* strndup wrapper */
 char *strndup(const char *s, size_t n)
{
	int len = strnlen(s, n);
	char *ret;
	ret = malloc(len+1);
	memcpy(ret, s, len);
	ret[len] = 0;
	return ret;
}

/* a free() call that zeros the pointer to prevent re-use */
void safe_free(void **ptr)
{
	if (!*ptr) {
		d_printf( "NULL free!\n");
		return;
	}
	free(*ptr);
	(*ptr) = NULL;
}


/* a vasprintf wrapper */
int vasprintf(char **strp, const char *fmt, va_list ap)
{
	int n;

	(*strp) = NULL;

	n = vsnprintf(NULL, 0, fmt, ap);
	if (n == 0) return 0;

	(*strp) = malloc(n+1);
	n = vsnprintf(*strp, n+1, fmt, ap);
	
	return n;
}

/* a asprintf wrapper */
int asprintf(char **strp, const char *fmt, ...)
{
	int n;
	va_list ap;

	va_start(ap, fmt);
	n = vasprintf(strp, fmt, ap);
	va_end(ap);

	return n;
}

/* a realloc wrapper */
void *realloc(void *ptr, size_t size)
{
	void *ret;

	if (!ptr) {
		return malloc(size);
	}

	if (size == 0) {
		free(ptr);
		return NULL;
	}

	ret = malloc(size);
	memcpy(ret, ptr, _malloc_size(ptr));
	free(ptr);
	return ret;
}
